from typing import Any

from apiron import JsonEndpoint
from munch import munchify

from pywisetransfer import Client
from pywisetransfer.base import Base


class UserService(Base):
    me = JsonEndpoint(path="/v1/me")
    get = JsonEndpoint(path="/v1/users/{userId}")


class User(object):
    def __init__(self, client: Client):
        self.service = UserService(client=client)

    def me(self) -> Any:
        return munchify(self.service.me())

    def get(self, user_id: str) -> Any:
        return munchify(self.service.get(user_id=user_id))



class WiseException(Exception):
    pass


class InvalidWebhookHeader(WiseException):
    pass


class InvalidWebhookRequest(WiseException):
    pass


class InvalidWebhookSignature(WiseException):
    pass



from typing import Any, Callable, Dict, Optional, Type

import apiron
from apiron.endpoint import JsonEndpoint
from functools import partial, update_wrapper, wraps
from requests.exceptions import HTTPError

from pywisetransfer.base import Base
from pywisetransfer.signing import sign_sca_challenge


class JsonEndpointWithSCA(JsonEndpoint):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.sca_headers: Dict[str, str] = {}

    # apiron's required_headers should be called override_headers
    @property
    def required_headers(self) -> Dict[str, str]:
        # It looks like 2FA only needs to be performed every few minutes,
        # but there seems to be no harm in re-sending the signature.
        return {**super().required_headers, **self.sca_headers}

    def __get__(self, instance: Optional[Base], owner: Type[Base]) -> Callable[..., Any]:
        caller = partial(apiron.client.call, owner, self)
        update_wrapper(caller, apiron.client.call)

        @wraps(apiron.client.call)
        def perform_2fa_if_needed(*args: Any, **kwargs: Any) -> Any:
            try:
                return caller(*args, **kwargs)
            except HTTPError as e:
                resp = e.response
                if (
                    resp.status_code == 403
                    and resp.headers["X-2FA-Approval-Result"] == "REJECTED"
                ):
                    challenge = resp.headers["X-2FA-Approval"]
                    if owner.client.private_key_data is None:  # type: ignore[union-attr]
                        raise Exception(
                            "Please provide pytransferwise.private_key_file or private_key_data to perform SCA authentication"
                        ) from e

                    self.sca_headers["X-Signature"] = sign_sca_challenge(
                        challenge, owner.client.private_key_data  # type: ignore[union-attr]
                    )
                    self.sca_headers["X-2FA-Approval"] = challenge
                    return caller(*args, **kwargs)
                raise

        return perform_2fa_if_needed



from typing import Any, List
from apiron import JsonEndpoint
from munch import munchify

from pywisetransfer import Client
from pywisetransfer.base import Base
from pywisetransfer.endpoint import JsonEndpointWithSCA


class BorderlessAccountService(Base):
    list = JsonEndpoint(path="/v1/borderless-accounts", required_params=["profileId"])
    statement = JsonEndpointWithSCA(
        path="/v3/profiles/{profile_id}/borderless-accounts/{account_id}/statement.json",
        required_params=["currency", "intervalStart", "intervalEnd"],
    )


class BorderlessAccount(object):
    def __init__(self, client: Client):
        self.service = BorderlessAccountService(client=client)

    def list(self, profile_id: str) -> List[Any]:
        accounts: List[Any] = self.service.list(params={"profileId": profile_id})
        return munchify(accounts)

    def statement(self, profile_id: str, account_id: str, currency: str, interval_start: str, interval_end: str) -> Any:
        return munchify(
            self.service.statement(
                profile_id=profile_id,
                account_id=account_id,
                params={
                    "currency": currency,
                    "intervalStart": interval_start,
                    "intervalEnd": interval_end,
                },
            )
        )



from typing import Any, List, Optional

from apiron import JsonEndpoint
from munch import munchify

from pywisetransfer import Client
from pywisetransfer.base import Base


class ProfileService(Base):
    list = JsonEndpoint(path="/v1/profiles")
    get = JsonEndpoint(path="/v1/profiles/{profile_id}")


class Profile(object):
    def __init__(self, client: Client):
        self.service = ProfileService(client=client)

    def list(self, type: Optional[str] = None) -> List[Any]:
        profiles: List[Any] = munchify(self.service.list())
        if type is None:
            return profiles
        return [p for p in profiles if p.type == type]

    def get(self, profile_id: str) -> Any:
        return munchify(self.service.get(profile_id=profile_id))



from typing import Any, Dict, Optional

from apiron import Service

from pywisetransfer import Client


class Base(Service):

    client: Optional[Client] = None

    def __init__(self, *args: Any, **kwargs: Any):
        # TODO: It would be nice to make 'client' a required argument here
        # Currently apiron performs a zero-argument constructor call to the
        # service base class to retrieve the 'required_headers' property,
        # so we cater for empty arguments here
        # https://github.com/ithaka/apiron/blob/v5.1.0/src/apiron/service/base.py#L7
        if "client" in kwargs and Base.client != kwargs["client"]:
            Base.client = kwargs["client"]
        if Base.client and Base.client.environment == "live":
            Base.domain = "https://api.transferwise.com"
        else:
            Base.domain = "https://api.sandbox.transferwise.tech"

    @property
    def required_headers(self) -> Dict[str, str]:  # type: ignore[override]
        if self.client:
            return {"Authorization": f"Bearer {self.client.api_key}"}
        return {}



WEBHOOK_SIGNATURE_PUBLIC_KEY_LIVE = b"""
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAvO8vXV+JksBzZAY6GhSO
XdoTCfhXaaiZ+qAbtaDBiu2AGkGVpmEygFmWP4Li9m5+Ni85BhVvZOodM9epgW3F
bA5Q1SexvAF1PPjX4JpMstak/QhAgl1qMSqEevL8cmUeTgcMuVWCJmlge9h7B1CS
D4rtlimGZozG39rUBDg6Qt2K+P4wBfLblL0k4C4YUdLnpGYEDIth+i8XsRpFlogx
CAFyH9+knYsDbR43UJ9shtc42Ybd40Afihj8KnYKXzchyQ42aC8aZ/h5hyZ28yVy
Oj3Vos0VdBIs/gAyJ/4yyQFCXYte64I7ssrlbGRaco4nKF3HmaNhxwyKyJafz19e
HwIDAQAB
-----END PUBLIC KEY-----
"""

WEBHOOK_SIGNATURE_PUBLIC_KEY_SANDBOX = b"""
-----BEGIN PUBLIC KEY-----
MIIBIjANBgkqhkiG9w0BAQEFAAOCAQ8AMIIBCgKCAQEAwpb91cEYuyJNQepZAVfP
ZIlPZfNUefH+n6w9SW3fykqKu938cR7WadQv87oF2VuT+fDt7kqeRziTmPSUhqPU
ys/V2Q1rlfJuXbE+Gga37t7zwd0egQ+KyOEHQOpcTwKmtZ81ieGHynAQzsn1We3j
wt760MsCPJ7GMT141ByQM+yW1Bx+4SG3IGjXWyqOWrcXsxAvIXkpUD/jK/L958Cg
nZEgz0BSEh0QxYLITnW1lLokSx/dTianWPFEhMC9BgijempgNXHNfcVirg1lPSyg
z7KqoKUN0oHqWLr2U1A+7kqrl6O2nx3CKs1bj1hToT1+p4kcMoHXA7kA+VBLUpEs
VwIDAQAB
-----END PUBLIC KEY-----
"""


def get_webhook_public_key(environment: str) -> bytes:
    return (
        WEBHOOK_SIGNATURE_PUBLIC_KEY_LIVE
        if environment == "live"
        else WEBHOOK_SIGNATURE_PUBLIC_KEY_SANDBOX
    )



from base64 import b64encode

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import load_pem_private_key


def sign_sca_challenge(challenge: str, private_key_data: bytes) -> str:
    private_key = load_pem_private_key(private_key_data, None)
    signature = private_key.sign(
        challenge.encode("ascii"), padding.PKCS1v15(), hashes.SHA256()
    )
    return b64encode(signature).decode("ascii")



from typing import Any

from apiron import JsonEndpoint
from munch import munchify

from pywisetransfer import Client
from pywisetransfer.base import Base


class SubscriptionService(Base):
    list = JsonEndpoint(path="/v3/profiles/{profile_id}/subscriptions")
    get = JsonEndpoint(path="/v3/profiles/{profile_id}/subscriptions/{subscription_id}")


class Subscription(object):
    def __init__(self, client: Client):
        self.service = SubscriptionService(client=client)

    def list(self, profile_id: str) -> Any:
        return munchify(self.service.list(profile_id=profile_id))

    def get(self, profile_id: str, subscription_id: str) -> Any:
        return munchify(
            self.service.get(profile_id=profile_id, subscription_id=subscription_id)
        )



from typing import Optional


class Client(object):
    def add_resources(self) -> None:
        from pywisetransfer.borderless_account import BorderlessAccount
        from pywisetransfer.profile import Profile
        from pywisetransfer.subscription import Subscription
        from pywisetransfer.user import User

        self.borderless_accounts = BorderlessAccount(client=self)
        self.profiles = Profile(client=self)
        self.subscriptions = Subscription(client=self)
        self.users = User(client=self)

    def __init__(
        self,
        api_key: str,
        environment: str = "sandbox",
        private_key_file: Optional[str] = None,
        private_key_data: Optional[bytes] = None,
    ):
        if api_key is None:
            raise KeyError("You must provide a value for pywisetransfer.api_key")

        if environment not in ("sandbox", "live"):
            raise KeyError("pywisetransfer.environment must be 'sandbox' or 'live'")

        if private_key_file is not None:
            if private_key_data is not None:
                raise ValueError(
                    "Please provide only one of pywisetransfer.private_key_file or private_key_data"
                )
            private_key_data = open(private_key_file, "rb").read()

        self.api_key = api_key
        self.environment = environment
        self.private_key_file = private_key_file
        self.private_key_data = private_key_data
        self.add_resources()



from base64 import b64decode

from pywisetransfer.exceptions import (
    InvalidWebhookHeader,
    InvalidWebhookRequest,
    InvalidWebhookSignature,
)
from pywisetransfer.keys import get_webhook_public_key

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import load_pem_public_key


def validate_request(request: "flask.Request", environment: str = "sandbox") -> None:
    if request.json is None:
        raise InvalidWebhookRequest("Webhook request does not contain JSON")

    if "X-Signature-SHA256" not in request.headers:
        raise InvalidWebhookRequest("Webhook request does not include SHA-256 signature")

    try:
        signature = b64decode(request.headers["X-Signature-SHA256"])
    except Exception:
        raise InvalidWebhookHeader("Cannot decode webhook signature")

    if not verify_signature(
        payload=request.data,
        signature=signature,
        environment=environment,
    ):
        raise InvalidWebhookSignature("Invalid webhook signature")


def verify_signature(payload: bytes, signature: bytes, environment: str = "sandbox") -> None:
    key_data = get_webhook_public_key(environment)
    public_key = load_pem_public_key(key_data, backend=default_backend())
    try:
        public_key.verify(signature, payload, padding.PKCS1v15(), hashes.SHA256())
        return True
    except InvalidSignature:
        return False
